// 極楽画像計測アプリケーション（Gokuraku_measure）
// Created by Tatsuya Shirai
// National Institute of Technology, Suzuka college
// Mechanical Department
//
// Ver. 1.0 : 2022.06.07
// Ver. 1.1 : 2022.06.08 : filter（ヘルプ表示にバグあり：動作に支障なし）
// Ver. 1.2 : 2022.06.08 : 縮小可能に．軽微なバグの修正（ヘルプ表示）
// Ver. 1.3 : 2022.06.08 : データフォーマットの変更（データファイル１行目のScale:の後にタブを追加）
// Ver. 1.4 : 2022.06.10 : データフォーマットの変更（データファイル１行目をscale, org_weight, org_heigtのタブ区切りに変更），ショートメッセージ表示機能の追加
// Ver. 2.0 : 2022.06.21 : ファイル名（パス無し）を記録する
// Ver. 2.1 : 2022.06.21 : 全てD&Dに対応，動的メモリー確保に統一，VisualizerがStick pictureに対応

CurrentImage ci;
ExportFile   ex_file;
ShortMessage shortMessage;
InputInteger inputInteger;

final int Scr_w = 600;
final int Scr_h = 550;

enum Menu_status {main, setting, sizeinputing, maxinputing, linewidthinputing};
Menu_status menu_status;

void setup()
{
  ci           = new CurrentImage();
  ex_file      = new ExportFile();
  shortMessage = new ShortMessage();
  inputInteger = new InputInteger();
  menu_status  = Menu_status.main;

  //サイズ変更を許可する
  surface.setResizable(true);

  drop_init();
}

void draw()
{
  int x  =   8;
  int y  =  48;
  int dx = 150;
  int dy =  40;

  noStroke();  
  fill(72, 72, 72);

  if ((ci.img != null) || (ci.folder_mode && ex_file.recording_now())) {
    // 画像が正常にロードされている場合
    int cnt = 0;
    rect(0, 0, width, height);

    // 画像表示
    if (! ex_file.recording_now()) image(ci.img, 0, 0);
    // 全プロット表示
    ci.plot_all();
    y = height - 16;
  } else {
    // メインメニュー：画像が選択されていない場合
    surface.setSize(Scr_w, Scr_h);
    rect(0, 0, Scr_w - 1, Scr_h - 1);
    fill(255, 125, 125);
    textSize(28);
    text("Gokuraku measure (@tatsuva) Ver." + Version, x, y); y += dy;
    textSize(24);
    fill(ci.plotMark.red(), ci.plotMark.green(), ci.plotMark.blue());
    if (menu_status == Menu_status.sizeinputing) {
      text("Plot size = " + inputInteger.value,   x, y); y += dy;
    } else if (menu_status == Menu_status.maxinputing) {
      text("max measure = " + inputInteger.value, x, y); y += dy;
    } else if (menu_status == Menu_status.linewidthinputing) {
      text("Line width = " + inputInteger.value, x, y);  y += dy;
    } else if (menu_status == Menu_status.main) {
      text("[I]", x, y);           text(": select image file",           x + dx, y); y += dy;
      text("[F]", x, y);           text(": select folder (folder mode)", x + dx, y); y += dy;
      text("[S]", x, y);           text(": setting submenu",         x + dx, y); y += dy;
      text("(Left click)",  x, y); text(": measure",                     x + dx, y); y += dy;
      text("Shift+(left)",  x, y); text(": back",                        x + dx, y); y += dy;
      text("Ctrl+(left)",   x, y); text(": skip",                        x + dx, y); y += dy;
      text("(Right clisk)", x, y); text(": end / select next image file",x + dx, y); y += dy; 
      text("[ESC]", x, y);         text(": end / exit",                  x + dx, y); y += dy;
      text("[+][-][0]", x, y);     text(": scale up/down/reset (=" + floorN((float)ci.scale, 4) + ")", x + dx, y); y += dy;
      String str = "None";
      if (ci.filter_no == 0) str = "Gray"; else if (ci.filter_no == 1) str = "Invert";  // none : filter_no == -1 
      text("[G]", x, y);           text(": change filter (=" + str + ")",x + dx, y); y += dy;
      text("[D]", x, y);           text(": show data (Console only)",    x + dx, y); y += dy;
    } else if (menu_status == Menu_status.setting) {
      dx = 64;
      text("[M]", x, y); text(": change max measure number (=" + ci.max_measure + ")",  x + dx, y); y += dy;
      text("[T]", x, y); text(": change plot type  (=" + ci.plotMark.get_shape() + ")", x + dx, y); y += dy;
      text("[C]", x, y); text(": change plot color (=" + ci.plotMark.color_str() + ")", x + dx, y); y += dy;
      text("[P]", x, y); text(": change plot size  (=" + ci.plotMark.size + ")",        x + dx, y); y += dy;
      text("[W]", x, y); text(": change line width (=" + ci.plotMark.lineWidth + ")",   x + dx, y); y += dy;
      text("[Q]", x, y); text(": quit sub menu",                                        x + dx, y); y += dy;
      y += 8;
  
      textSize(16);
      fill(255, 255, 255);
      if (ci.max_measure == 0) text("Max measure number = 0, no limit",       x, y);
      else                     text("Max measure number = " + ci.max_measure, x, y);
      y += 20;
    }
  }
  shortMessage.display_short_message(x, y);
}

void mousePressed()
{
  if (mouseButton == RIGHT) {    // 右クリック時
    ex_file.flush_data();  // データを出力
    if (ci.folder_mode) {
      // フォルダーモードの場合の
      if (ci.next_image() == null) ci.quit_folder_mode (); // もう画像がフォルダー内に無いならば終了
    } else {
      // 単体画像モードの場合はファイル選択ダイアログボックスを表示
      open_fileselector();
    }
  } else if (mouseButton == LEFT) {  // 左クリック時
    if (ci.img != null) {
      if (keyPressed && keyCode == SHIFT)        ci.back();                   // 計測のキャンセル（バック）
      else if (keyPressed && keyCode == CONTROL) ci.skip();                   // 計測のスキップ
      else                                       ci.measure(mouseX, mouseY);  // 計測
    }
  }
}

void keyPressed()
{
  if (menu_status == Menu_status.sizeinputing) {  // plot sizeを修正時
    // plot_size を修正時
    if (inputInteger.inputing_now) {
      if (key == ENTER) {
        int num = inputInteger.value;
        if (num > 0) ci.plotMark.size = num;
        inputInteger.end();
        menu_status = Menu_status.setting;
        shortMessage.set_message("Change plot size to " + ci.plotMark.size);
      } else if (key == ESC) {
        menu_status = Menu_status.setting;
        shortMessage.set_message("Canceled");
        key = 0;
      } else inputInteger.keyin(key);
      return;
    }
  } else if (menu_status == Menu_status.maxinputing) {  // max_measure を修正時
    if (inputInteger.inputing_now) {
      if (key == ENTER) {
        int num = inputInteger.value;
        if (num >= 0) ci.max_measure = num;
        inputInteger.end();
        menu_status = Menu_status.setting;
        shortMessage.set_message("Change max measure to " + ci.max_measure);
      } else if (key == ESC) {
        menu_status = Menu_status.setting;
        shortMessage.set_message("Canceled");
        key = 0;
      } else inputInteger.keyin(key);
      return;
    }
  } else if (menu_status == Menu_status.linewidthinputing) {
    // lineWidth を修正時
    if (inputInteger.inputing_now) {
      if (key == ENTER) {
        int num = inputInteger.value;
        if (num >= 1) ci.plotMark.lineWidth = num;
        inputInteger.end();
        menu_status = Menu_status.setting;
        shortMessage.set_message("Change line width to " + ci.plotMark.lineWidth);
      } else if (key == ESC) {
        menu_status = Menu_status.setting;
        shortMessage.set_message("Canceled");
        key = 0;
      } else inputInteger.keyin(key);
      return;
    }
  } else if (menu_status == Menu_status.main) {
    switch (keyCode) {
      case 'I' :
        open_fileselector();
        break;
      case 'F' :
        File fp;
        fp = new File(ci.folder_mode_folder);
        selectFolder("Select a folder to process:", "folderSelected", fp);  
        break;
      case 'S' :
        // 設定のサブメニューに移動
        menu_status = Menu_status.setting;
        break;
      case ESC : 
        if (ci.img == null) terminate(); else {
          // 画像計測中
          ex_file.flush_data();                      // データを出力
          if (ci.folder_mode) ci.quit_folder_mode(); // フォルダーモードならば終了
          key = 0;                                   // keyをクリアしないとProcessingが終了してしまう
        }
        // 画像計測時以外は，このままProcessingが終了する
        break;
      case 'G' :
        ci.change_filter();
        String fstr = "none";
        if (ci.filter_no >= 0) fstr = filter_name[ci.filter_no];
        shortMessage.set_message("Change filter to [" + fstr + "]");
        break;
      case 107 :
      case  59 :
      case '+' :
        ci.rescale(Scale_rate);
        break;
      case 109 :
      case '-' :
        ci.rescale(1.0/Scale_rate); 
        break;
      case  96 :
      case '0' :
        ci.rescale( 0.0);
        break;
      case 'D' : 
        show_alldata();
        break;
    }
  } else if (menu_status == Menu_status.setting) {
    switch (keyCode) {
      case 'M' :
        inputInteger.start(ci.max_measure);
        menu_status = Menu_status.maxinputing;
        break;
      case 'T' : 
        ci.plotMark.change_shape();
        shortMessage.set_message("Change plot type to " + ci.plotMark.get_shape());
        break;
      case 'C' :
        ci.plotMark.change_color();
        shortMessage.set_message("Change plot color to " + ci.plotMark.color_str());
        break;
      case 'P' :
        inputInteger.start(ci.plotMark.size);
        menu_status = Menu_status.sizeinputing;
        break;
      case 'W' :
        inputInteger.start(ci.plotMark.lineWidth);
        menu_status = Menu_status.linewidthinputing;
        break;
      case 'Q' :
        menu_status = Menu_status.main;
        break;
    }
    key = 0;
  }
}

//// ファイル・フォルダー選択ダイアログボックス関係の関数
// ファイル選択ダイアログボックスの呼び出し
void open_fileselector()
{
  selectInput("Select a file to process:", "fileSelected");
}
// ファイル選択後の処理
void fileSelected(File selection) 
{
  ex_file.flush_data();
  if (selection == null) ci.img = null;
  else ci.load(selection.getAbsolutePath());
}

// フォルダーモード
void folderSelected(File selection) {
  ci.folderSelected(selection);          // selectFolder()から直接呼び出すことができないので．
}

//// それ以外のユーザー関数
// 終了時の処理
void terminate()
{
  ex_file.flush_data();  // 計測済みならば出力
  println("Exit program.");
  exit();
}

// 計測済みのデータをコンソールに表示（デバッグ用）
void show_alldata()
{
  if (ci.img == null) return;
  println("[Data]");
  if (ci._counter <= 0) println("no data");
  else for (int n = 0; n < ci._counter; n++) println(n + " : " + ci.measured[n].px + " , " + ci.measured[n].py); 
}

//// ドラッグ＆ドロップ関係
// 上記URL参照
// ドロップされたファイルの処理
void fileSelected(List<File> fs) {
  File [] fileList = {};
  File oneDirectory = null;  // 特例のための変数
  for (File f:fs) {
    // ドラッグ＆ドロップされたリストの中にフォルダーが含まれていた場合は無視する
    if (f.isDirectory()) {
      oneDirectory = f;
      continue;
    }
    fileList = (File [])append(fileList, f);  // ListからArrayへ変更
  }
  if (fileList.length == 0) {
    // 特例として，D&Dで単一のフォルダーのみがDropされた時はフォルダーモードとして動作する
    if (oneDirectory != null) {
      ci.folderSelected(oneDirectory);
      return;
    }
    return;  // ファイルが一つも存在しなかった場合
  }
  ci.open_files(fileList);
}

import java.awt.datatransfer.*;  
import java.awt.dnd.*;  
import java.io.File;  
import java.io.IOException;  
import java.awt.Component;
import java.util.List;

DropTarget dropTarget;
Component  component;

void drop_init() {  
  component = (Component)this.surface.getNative();
  dropTarget = new DropTarget(component, new DropTargetListener() {
    public void dragEnter(DropTargetDragEvent dtde) {
    }  
    public void dragOver(DropTargetDragEvent dtde) {
    }  
    public void dropActionChanged(DropTargetDragEvent dtde) {
    }  
    public void dragExit(DropTargetEvent dte) {
    }  
    public void drop(DropTargetDropEvent dtde) {  
      dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);  
      Transferable trans = dtde.getTransferable();  
      List<File> fileNameList = null;  
      if (trans.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {  
        try {  
          fileNameList = (List<File>)  
            trans.getTransferData(DataFlavor.javaFileListFlavor);
        } 
        catch (UnsupportedFlavorException ex) {
        } 
        catch (IOException ex) {
        }
      }  
      if (fileNameList == null) return;  
      fileSelected(fileNameList);
    }
  }
  );
}
